"""This module contains everything related to split.io exceptions"""
from __future__ import absolute_import, division, print_function, unicode_literals

from splitio.client.factory import TimeoutException
from splitio.storage.adapters.redis import SentinelConfigurationException
