from __future__ import absolute_import, division, print_function, \
    unicode_literals

from splitio.client.factory import get_factory
from splitio.client.key import Key
from splitio.version import __version__
